use http::Uri;
use serde::Deserialize;
use thiserror::Error;

pub mod auth;
pub mod user;
mod util;

const DEFAULT_LAST_FM_API_URI: &str = "https://ws.audioscrobbler.com/2.0/";

#[derive(Debug, Clone)]
pub struct LastfmClient {
    api_url: String,
    key: String,
    secret: String,
}

impl LastfmClient {
    pub fn new(key: &str, secret: &str) -> Self {
        LastfmClient {
            api_url: DEFAULT_LAST_FM_API_URI.to_string(),
            key: key.to_string(),
            secret: secret.to_string(),
        }
    }
    pub fn new_with_custom_api(url: &str, key: &str, secret: &str) -> Self {
        LastfmClient {
            api_url: url.parse::<Uri>().expect("Invalid API URL").to_string(),
            key: key.to_string(),
            secret: secret.to_string(),
        }
    }

    fn sign_call(&self, method: &str, params: &[(&str, &str)]) -> md5::Digest {
        let mut string = String::new();

        string.push_str("api_key");
        string.push_str(self.key.as_str());

        string.push_str("method");
        string.push_str(method);

        for i in params {
            string.push_str(&i.0);
            string.push_str(&i.1);
        }

        string.push_str(self.secret.as_str());

        md5::compute(string)
    }

    fn build_get_url(&self, method: &str, params: &[(&str, &str)], sign: bool) -> String {
        let mut api_url = format!("{}?method={}&api_key={}", self.api_url, method, self.key);

        for i in params {
            api_url.push_str(format!("&{}={}", i.0, i.1).as_str());
        }

        if sign {
            api_url.push_str(format!("&api_sig={:x}", self.sign_call(method, params)).as_str());
        }

        api_url.push_str("&format=json");

        api_url
    }
}

#[derive(Debug, Error)]
#[repr(u16)]
pub enum LastfmApiError {
    #[error("This service does not exist")]
    InvalidService = 2,
    #[error("No method with that name in this package")]
    InvalidMethod,
    #[error("You do not have permissions to access the service")]
    AuthenticationFailed,
    #[error("This services doesn't exist in that format")]
    InvalidFormat,
    #[error("Your request is missing a required parameter")]
    InvalidParameters,
    #[error("Invalid resource specified")]
    InvalidResource,
    #[error("Something else went wrong")]
    OperationFailed,
    #[error("Please re-authenticate")]
    InvalidSessionKey,
    #[error("You must be granted a valid key by last.fm")]
    InvalidAPIKey,
    #[error("This service is temporarily offline. Try again later.")]
    ServiceOffline,
    #[error("Invalid method signature supplied")]
    InvalidMethodSignature = 13,
    #[error("This token has not been authorized")]
    UnauthorizedToken,
    #[error("This token has expired")]
    ExpiredToken,
    #[error("There was a temporary error processing your request. Please try again")]
    TemporaryError,
    #[error("Access for your account has been suspended, please contact Last.fm")]
    SuspendedKey = 26,
    #[error("Your IP has made too many requests in a short period")]
    RateLimitExceeded = 29,
    #[error("Unknown Error")]
    Unknown,
    #[error(transparent)]
    Reqwest(#[from] reqwest::Error),
    #[error(transparent)]
    SerdeJson(#[from] serde_json::Error),
}

#[derive(Deserialize, Debug)]
struct ApiError {
    error: u16,
    message: String,
}

impl ApiError {
    pub fn into_error(self) -> LastfmApiError {
        self.error.into()
    }
}

impl From<u16> for LastfmApiError {
    fn from(error: u16) -> Self {
        match error {
            2 => LastfmApiError::InvalidService,
            3 => LastfmApiError::InvalidMethod,
            4 => LastfmApiError::AuthenticationFailed,
            5 => LastfmApiError::InvalidFormat,
            6 => LastfmApiError::InvalidParameters,
            7 => LastfmApiError::InvalidResource,
            8 => LastfmApiError::OperationFailed,
            9 => LastfmApiError::InvalidSessionKey,
            10 => LastfmApiError::InvalidAPIKey,
            11 => LastfmApiError::ServiceOffline,
            13 => LastfmApiError::InvalidMethodSignature,
            14 => LastfmApiError::UnauthorizedToken,
            15 => LastfmApiError::ExpiredToken,
            26 => LastfmApiError::SuspendedKey,
            29 => LastfmApiError::RateLimitExceeded,
            _ => LastfmApiError::Unknown,
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;
}
